@@ -1,26 +1,6 @@
-/* Spa
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <errno.h>
 #include <string.h>
@@ -29,6 +9,7 @@
 #include <spa/support/plugin.h>
 #include <spa/support/log.h>
 #include <spa/support/cpu.h>
+#include <spa/support/loop.h>
 #include <spa/utils/list.h>
 #include <spa/utils/names.h>
 #include <spa/utils/string.h>
@@ -42,12 +23,16 @@
 #include "mix-ops.h"
 
 #undef SPA_LOG_TOPIC_DEFAULT
-#define SPA_LOG_TOPIC_DEFAULT log_topic
-static struct spa_log_topic *log_topic = &SPA_LOG_TOPIC(0, "spa.audiomixer");
+#define SPA_LOG_TOPIC_DEFAULT &log_topic
+SPA_LOG_TOPIC_DEFINE_STATIC(log_topic, "spa.audiomixer");
+
+#define DEFAULT_RATE		48000
+#define DEFAULT_CHANNELS	2
 
 #define MAX_BUFFERS     64
-#define MAX_PORTS       128
+#define MAX_PORTS       512
 #define MAX_CHANNELS    64
+#define MAX_ALIGN	MIX_OPS_MAX_ALIGN
 
 #define PORT_DEFAULT_VOLUME	1.0
 #define PORT_DEFAULT_MUTE	false
@@ -80,7 +65,7 @@ struct port {
 
 	struct port_props props;
 
-	struct spa_io_buffers *io;
+	struct spa_io_buffers *io[2];
 
 	uint64_t info_all;
 	struct spa_port_info info;
@@ -103,6 +88,10 @@ struct impl {
 	struct spa_log *log;
 	struct spa_cpu *cpu;
 	uint32_t cpu_flags;
+	uint32_t max_align;
+
+	struct spa_loop *data_loop;
+
 	uint32_t quantum_limit;
 
 	struct mix_ops ops;
@@ -111,6 +100,8 @@ struct impl {
 	struct spa_node_info info;
 	struct spa_param_info params[8];
 
+	struct spa_io_position *position;
+
 	struct spa_hook_list hooks;
 
 	uint32_t port_count;
@@ -118,6 +109,9 @@ struct impl {
 	struct port *in_ports[MAX_PORTS];
 	struct port out_ports[1];
 
+	struct buffer *mix_buffers[MAX_PORTS];
+	const void *mix_datas[MAX_PORTS];
+
 	int n_formats;
 	struct spa_audio_info format;
 
@@ -128,13 +122,16 @@ struct impl {
 };
 
 #define PORT_VALID(p)                ((p) != NULL && (p)->valid)
+#define CHECK_ANY_IN(this,d,p)       ((d) == SPA_DIRECTION_INPUT && (p) == SPA_ID_INVALID)
 #define CHECK_FREE_IN_PORT(this,d,p) ((d) == SPA_DIRECTION_INPUT && (p) < MAX_PORTS && !PORT_VALID(this->in_ports[(p)]))
 #define CHECK_IN_PORT(this,d,p)      ((d) == SPA_DIRECTION_INPUT && (p) < MAX_PORTS && PORT_VALID(this->in_ports[(p)]))
 #define CHECK_OUT_PORT(this,d,p)     ((d) == SPA_DIRECTION_OUTPUT && (p) == 0)
 #define CHECK_PORT(this,d,p)         (CHECK_OUT_PORT(this,d,p) || CHECK_IN_PORT (this,d,p))
+#define CHECK_PORT_ANY(this,d,p)     (CHECK_ANY_IN(this,d,p) || CHECK_PORT(this,d,p))
 #define GET_IN_PORT(this,p)          (this->in_ports[p])
 #define GET_OUT_PORT(this,p)         (&this->out_ports[p])
 #define GET_PORT(this,d,p)           (d == SPA_DIRECTION_INPUT ? GET_IN_PORT(this,p) : GET_OUT_PORT(this,p))
+#define GET_PORT_ANY(this,d,p)       (CHECK_ANY_IN(this,d,p) ? NULL : GET_PORT(this,d,p))
 
 static int impl_node_enum_params(void *object, int seq,
 			uint32_t id, uint32_t start, uint32_t num,
@@ -151,7 +148,16 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 
 static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 {
-	return -ENOTSUP;
+	struct impl *this = object;
+
+	switch (id) {
+	case SPA_IO_Position:
+		this->position = data;
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	return 0;
 }
 
 static int impl_node_send_command(void *object, const struct spa_command *command)
@@ -315,8 +321,7 @@ impl_node_remove_port(void *object, enum spa_direction direction, uint32_t port_
 	return 0;
 }
 
-static int port_enum_formats(void *object,
-			     enum spa_direction direction, uint32_t port_id,
+static int port_enum_formats(void *object, struct port *port,
 			     uint32_t index,
 			     struct spa_pod **param,
 			     struct spa_pod_builder *builder)
@@ -351,8 +356,10 @@ static int port_enum_formats(void *object,
 								SPA_AUDIO_FORMAT_U24_32,
 								SPA_AUDIO_FORMAT_F32,
 								SPA_AUDIO_FORMAT_F64),
-				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(44100, 1, INT32_MAX),
-				SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(2, 1, INT32_MAX));
+				SPA_FORMAT_AUDIO_rate,     SPA_POD_CHOICE_RANGE_Int(
+								DEFAULT_RATE, 1, INT32_MAX),
+				SPA_FORMAT_AUDIO_channels, SPA_POD_CHOICE_RANGE_Int(
+								DEFAULT_CHANNELS, 1, INT32_MAX));
 		}
 		break;
 	default:
@@ -378,9 +385,9 @@ impl_node_port_enum_params(void *object, int seq,
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 	spa_return_val_if_fail(num != 0, -EINVAL);
-	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
+	spa_return_val_if_fail(CHECK_PORT_ANY(this, direction, port_id), -EINVAL);
 
-	port = GET_PORT(this, direction, port_id);
+	port = GET_PORT_ANY(this, direction, port_id);
 
 	result.id = id;
 	result.next = start;
@@ -391,12 +398,12 @@ impl_node_port_enum_params(void *object, int seq,
 
 	switch (id) {
 	case SPA_PARAM_EnumFormat:
-		if ((res = port_enum_formats(this, direction, port_id, result.index, &param, &b)) <= 0)
+		if ((res = port_enum_formats(this, port, result.index, &param, &b)) <= 0)
 			return res;
 		break;
 
 	case SPA_PARAM_Format:
-		if (!port->have_format)
+		if (port == NULL || !port->have_format)
 			return -EIO;
 		if (result.index > 0)
 			return 0;
@@ -405,7 +412,7 @@ impl_node_port_enum_params(void *object, int seq,
 		break;
 
 	case SPA_PARAM_Buffers:
-		if (!port->have_format)
+		if (port == NULL || !port->have_format)
 			return -EIO;
 		if (result.index > 0)
 			return 0;
@@ -440,6 +447,12 @@ impl_node_port_enum_params(void *object, int seq,
 				SPA_PARAM_IO_id,   SPA_POD_Id(SPA_IO_Buffers),
 				SPA_PARAM_IO_size, SPA_POD_Int(sizeof(struct spa_io_buffers)));
 			break;
+		case 1:
+			param = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_ParamIO, id,
+				SPA_PARAM_IO_id,   SPA_POD_Id(SPA_IO_AsyncBuffers),
+				SPA_PARAM_IO_size, SPA_POD_Int(sizeof(struct spa_io_async_buffers)));
+			break;
 		default:
 			return 0;
 		}
@@ -535,6 +548,8 @@ static int port_set_format(void *object,
 
 	port = GET_PORT(this, direction, port_id);
 
+	spa_return_val_if_fail(!this->started || port->io == NULL, -EIO);
+
 	if (format == NULL) {
 		if (port->have_format) {
 			port->have_format = false;
@@ -559,6 +574,10 @@ static int port_set_format(void *object,
 			if (memcmp(&info, &this->format, sizeof(struct spa_audio_info)))
 				return -EINVAL;
 		} else {
+			if (info.info.raw.format == 0 ||
+			    info.info.raw.channels == 0)
+				return -EINVAL;
+
 			this->ops.fmt = info.info.raw.format;
 			this->ops.n_channels = info.info.raw.channels;
 			this->ops.cpu_flags = this->cpu_flags;
@@ -638,10 +657,15 @@ impl_node_port_use_buffers(void *object,
 
 	port = GET_PORT(this, direction, port_id);
 
-	spa_return_val_if_fail(port->have_format, -EIO);
+	spa_return_val_if_fail(!this->started || port->io == NULL, -EIO);
 
 	clear_buffers(this, port);
 
+	if (n_buffers > 0 && !port->have_format)
+		return -EIO;
+	if (n_buffers > MAX_BUFFERS)
+		return -ENOSPC;
+
 	for (i = 0; i < n_buffers; i++) {
 		struct buffer *b;
 		struct spa_data *d = buffers[i]->datas;
@@ -658,7 +682,7 @@ impl_node_port_use_buffers(void *object,
 				      buffers[i]);
 			return -EINVAL;
 		}
-		if (!SPA_IS_ALIGNED(d[0].data, 16)) {
+		if (!SPA_IS_ALIGNED(d[0].data, this->max_align)) {
 			spa_log_warn(this->log, "%p: memory on buffer %d not aligned", this, i);
 		}
 		if (direction == SPA_DIRECTION_OUTPUT)
@@ -673,6 +697,30 @@ impl_node_port_use_buffers(void *object,
 	return 0;
 }
 
+struct io_info {
+	struct port *port;
+	void *data;
+	size_t size;
+};
+
+static int do_port_set_io(struct spa_loop *loop, bool async, uint32_t seq,
+		const void *data, size_t size, void *user_data)
+{
+	struct io_info *info = user_data;
+	if (info->size >= sizeof(struct spa_io_async_buffers)) {
+		struct spa_io_async_buffers *ab = info->data;
+		info->port->io[0] = &ab->buffers[info->port->direction];
+		info->port->io[1] = &ab->buffers[info->port->direction^1];
+	} else if (info->size >= sizeof(struct spa_io_buffers)) {
+		info->port->io[0] = info->data;
+		info->port->io[1] = info->data;
+	} else {
+		info->port->io[0] = NULL;
+		info->port->io[1] = NULL;
+	}
+	return 0;
+}
+
 static int
 impl_node_port_set_io(void *object,
 		      enum spa_direction direction, uint32_t port_id,
@@ -680,6 +728,7 @@ impl_node_port_set_io(void *object,
 {
 	struct impl *this = object;
 	struct port *port;
+	struct io_info info;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
@@ -689,10 +738,15 @@ impl_node_port_set_io(void *object,
 	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
 
 	port = GET_PORT(this, direction, port_id);
+	info.port = port;
+	info.data = data;
+	info.size = size;
 
 	switch (id) {
 	case SPA_IO_Buffers:
-		port->io = data;
+	case SPA_IO_AsyncBuffers:
+		spa_loop_invoke(this->data_loop,
+                               do_port_set_io, SPA_ID_INVALID, NULL, 0, true, &info);
 		break;
 	default:
 		return -ENOENT;
@@ -724,12 +778,13 @@ static int impl_node_process(void *object)
 	struct buffer **buffers;
 	struct buffer *outb;
 	const void **datas;
+	uint32_t cycle = this->position->clock.cycle & 1;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
 	outport = GET_OUT_PORT(this, 0);
-	outio = outport->io;
-	spa_return_val_if_fail(outio != NULL, -EIO);
+	if ((outio = outport->io[cycle]) == NULL)
+		return -EIO;
 
 	spa_log_trace_fp(this->log, "%p: status %p %d %d",
 			this, outio, outio->status, outio->buffer_id);
@@ -743,9 +798,9 @@ static int impl_node_process(void *object)
 		outio->buffer_id = SPA_ID_INVALID;
 	}
 
-	buffers = alloca(MAX_PORTS * sizeof(struct buffer *));
-        datas = alloca(MAX_PORTS * sizeof(void *));
-        n_buffers = 0;
+	buffers = this->mix_buffers;
+	datas = this->mix_datas;
+	n_buffers = 0;
 
 	maxsize = UINT32_MAX;
 
@@ -753,41 +808,52 @@ static int impl_node_process(void *object)
 		struct port *inport = GET_IN_PORT(this, i);
 		struct spa_io_buffers *inio = NULL;
 		struct buffer *inb;
+		struct spa_data *bd;
+		uint32_t size, offs;
 
-		if (SPA_UNLIKELY(!PORT_VALID(inport) ||
-		    (inio = inport->io) == NULL ||
-		    inio->buffer_id >= inport->n_buffers ||
-		    inio->status != SPA_STATUS_HAVE_DATA)) {
-			spa_log_trace_fp(this->log, "%p: skip input idx:%d valid:%d "
+		if (SPA_UNLIKELY(!PORT_VALID(inport) || (inio = inport->io[cycle]) == NULL)) {
+			spa_log_trace_fp(this->log, "%p: skip input idx:%d valid:%d io:%p/%p/%d",
+					this, i, PORT_VALID(inport),
+					inport->io[0], inport->io[1], cycle);
+			continue;
+		}
+		if (inio->buffer_id >= inport->n_buffers ||
+		    inio->status != SPA_STATUS_HAVE_DATA) {
+			spa_log_trace_fp(this->log, "%p: skip input idx:%d "
 					"io:%p status:%d buf_id:%d n_buffers:%d", this,
-				i, PORT_VALID(inport), inio,
-				inio ? inio->status : -1,
-				inio ? inio->buffer_id : SPA_ID_INVALID,
-				inport->n_buffers);
+				i, inio, inio->status, inio->buffer_id, inport->n_buffers);
 			continue;
 		}
 
 		inb = &inport->buffers[inio->buffer_id];
-		maxsize = SPA_MIN(inb->buffer->datas[0].chunk->size, maxsize);
+		bd = &inb->buffer->datas[0];
 
-		spa_log_trace_fp(this->log, "%p: mix input %d %p->%p %d %d %d", this,
-				i, inio, outio, inio->status, inio->buffer_id, maxsize);
+		offs = SPA_MIN(bd->chunk->offset, bd->maxsize);
+		size = SPA_MIN(bd->maxsize - offs, bd->chunk->size);
+		maxsize = SPA_MIN(size, maxsize);
 
-		datas[n_buffers] = inb->buffer->datas[0].data;
-		buffers[n_buffers++] = inb;
+		spa_log_trace_fp(this->log, "%p: mix input %d %p->%p %d %d %d:%d/%d", this,
+				i, inio, outio, inio->status, inio->buffer_id,
+				offs, size, this->stride);
+
+		if (!SPA_FLAG_IS_SET(bd->chunk->flags, SPA_CHUNK_FLAG_EMPTY)) {
+			datas[n_buffers] = SPA_PTROFF(bd->data, offs, void);
+			buffers[n_buffers++] = inb;
+		}
 		inio->status = SPA_STATUS_NEED_DATA;
 	}
 
 	outb = dequeue_buffer(this, outport);
         if (SPA_UNLIKELY(outb == NULL)) {
-                spa_log_trace(this->log, "%p: out of buffers", this);
+		if (outport->n_buffers > 0)
+			spa_log_warn(this->log, "%p: out of buffers (%d)", this,
+					outport->n_buffers);
                 return -EPIPE;
         }
 
 	if (n_buffers == 1) {
 		*outb->buffer = *buffers[0]->buffer;
-	}
-	else {
+	} else {
 		struct spa_data *d = outb->buf.datas;
 
 		*outb->buffer = outb->buf;
@@ -797,6 +863,7 @@ static int impl_node_process(void *object)
 		d[0].chunk->offset = 0;
 		d[0].chunk->size = maxsize;
 		d[0].chunk->stride = this->stride;
+		SPA_FLAG_UPDATE(d[0].chunk->flags, SPA_CHUNK_FLAG_EMPTY, n_buffers == 0);
 
 		mix_ops_process(&this->ops, d[0].data,
 				datas, n_buffers, maxsize / this->stride);
@@ -885,11 +952,19 @@ impl_init(const struct spa_handle_factory *factory,
 	this = (struct impl *) handle;
 
 	this->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
-	spa_log_topic_init(this->log, log_topic);
+	spa_log_topic_init(this->log, &log_topic);
+
+	this->data_loop = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_DataLoop);
+	if (this->data_loop == NULL) {
+		spa_log_error(this->log, "a data loop is needed");
+		return -EINVAL;
+	}
 
 	this->cpu = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_CPU);
-	if (this->cpu)
+	if (this->cpu) {
 		this->cpu_flags = spa_cpu_get_flags(this->cpu);
+		this->max_align = SPA_MIN(MAX_ALIGN, spa_cpu_get_max_align(this->cpu));
+	}
 
 	for (i = 0; info && i < info->n_items; i++) {
 		const char *k = info->items[i].key;
