@@ -1,26 +1,7 @@
-/* Spa
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
+
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -30,7 +11,9 @@
 #include <sys/mman.h>
 #include <poll.h>
 
-static void v4l2_on_fd_events(struct spa_source *source);
+#include <spa/pod/dynamic.h>
+#include <spa/utils/cleanup.h>
+#include <spa/utils/result.h>
 
 static int xioctl(int fd, int request, void *arg)
 {
@@ -43,7 +26,6 @@ static int xioctl(int fd, int request, void *arg)
 	return err;
 }
 
-
 int spa_v4l2_open(struct spa_v4l2_device *dev, const char *path)
 {
 	struct stat st;
@@ -57,7 +39,7 @@ int spa_v4l2_open(struct spa_v4l2_device *dev, const char *path)
 		return -EIO;
 	}
 
-	spa_log_info(dev->log, "Playback device is '%s'", path);
+	spa_log_info(dev->log, "device is '%s'", path);
 
 	dev->fd = open(path, O_RDWR | O_NONBLOCK, 0);
 	if (dev->fd == -1) {
@@ -85,6 +67,7 @@ int spa_v4l2_open(struct spa_v4l2_device *dev, const char *path)
 		spa_log_error(dev->log, "'%s' QUERYCAP: %m", path);
 		goto error_close;
 	}
+	snprintf(dev->path, sizeof(dev->path), "%s", path);
 	return 0;
 
 error_close:
@@ -110,7 +93,7 @@ int spa_v4l2_close(struct spa_v4l2_device *dev)
 	if (dev->active || dev->have_format)
 		return 0;
 
-	spa_log_info(dev->log, "close");
+	spa_log_info(dev->log, "close '%s'", dev->path);
 
 	if (close(dev->fd))
 		spa_log_warn(dev->log, "close: %m");
@@ -137,7 +120,6 @@ static int spa_v4l2_buffer_recycle(struct impl *this, uint32_t buffer_id)
 		spa_log_error(this->log, "'%s' VIDIOC_QBUF: %m", this->props.device);
 		return -err;
 	}
-
 	return 0;
 }
 
@@ -164,6 +146,8 @@ static int spa_v4l2_clear_buffers(struct impl *this)
 		if (SPA_FLAG_IS_SET(b->flags, BUFFER_FLAG_MAPPED)) {
 			munmap(b->ptr, d[0].maxsize);
 		}
+		if (b->mmap_ptr)
+			munmap(b->mmap_ptr, b->v4l2_buffer.length);
 		if (SPA_FLAG_IS_SET(b->flags, BUFFER_FLAG_ALLOCATED)) {
 			spa_log_debug(this->log, "close %d", (int) d[0].fd);
 			close(d[0].fd);
@@ -351,11 +335,9 @@ static const struct format_info format_info[] = {
 
 static const struct format_info *fourcc_to_format_info(uint32_t fourcc)
 {
-	size_t i;
-
-	for (i = 0; i < SPA_N_ELEMENTS(format_info); i++) {
-		if (format_info[i].fourcc == fourcc)
-			return &format_info[i];
+	SPA_FOR_EACH_ELEMENT_VAR(format_info, i) {
+		if (i->fourcc == fourcc)
+			return i;
 	}
 	return NULL;
 }
@@ -363,11 +345,9 @@ static const struct format_info *fourcc_to_format_info(uint32_t fourcc)
 #if 0
 static const struct format_info *video_format_to_format_info(uint32_t format)
 {
-	int i;
-
-	for (i = 0; i < SPA_N_ELEMENTS(format_info); i++) {
-		if (format_info[i].format == format)
-			return &format_info[i];
+	SPA_FOR_EACH_ELEMENT_VAR(format_info, i) {
+		if (i->format == format)
+			return i;
 	}
 	return NULL;
 }
@@ -381,19 +361,20 @@ static const struct format_info *find_format_info_by_media_type(uint32_t type,
 	size_t i;
 
 	for (i = startidx; i < SPA_N_ELEMENTS(format_info); i++) {
-		if ((format_info[i].media_type == type) &&
-		    (format_info[i].media_subtype == subtype) &&
-		    (format == 0 || format_info[i].format == format))
-			return &format_info[i];
+		const struct format_info *fi = &format_info[i];
+		if ((fi->media_type == type) &&
+		    (fi->media_subtype == subtype) &&
+		    (format == 0 || fi->format == format))
+			return fi;
 	}
 	return NULL;
 }
 
-static uint32_t
+static int
 enum_filter_format(uint32_t media_type, int32_t media_subtype,
 		   const struct spa_pod *filter, uint32_t index)
 {
-	uint32_t video_format = 0;
+	uint32_t video_format = SPA_VIDEO_FORMAT_UNKNOWN;
 
 	switch (media_type) {
 	case SPA_MEDIA_TYPE_video:
@@ -405,11 +386,10 @@ enum_filter_format(uint32_t media_type, int32_t media_subtype,
 			const uint32_t *values;
 
 			if (!(p = spa_pod_find_prop(filter, NULL, SPA_FORMAT_VIDEO_format)))
-				return SPA_VIDEO_FORMAT_UNKNOWN;
+				return -ENOENT;
 
 			val = spa_pod_get_values(&p->value, &n_values, &choice);
-
-			if (val->type != SPA_TYPE_Id)
+			if (val->type != SPA_TYPE_Id || n_values == 0)
 				return SPA_VIDEO_FORMAT_UNKNOWN;
 
 			values = SPA_POD_BODY(val);
@@ -525,10 +505,11 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 	int res, n_fractions;
 	const struct format_info *info;
 	struct spa_pod_choice *choice;
-	uint32_t filter_media_type, filter_media_subtype, video_format;
+	uint32_t filter_media_type, filter_media_subtype;
 	struct spa_v4l2_device *dev = &port->dev;
 	uint8_t buffer[1024];
-	struct spa_pod_builder b = { 0 };
+	spa_auto(spa_pod_dynamic_builder) b = { 0 };
+	struct spa_pod_builder_state state;
 	struct spa_pod_frame f[2];
 	struct spa_result_node_params result;
 	uint32_t count = 0;
@@ -536,6 +517,9 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 	if ((res = spa_v4l2_open(dev, this->props.device)) < 0)
 		return res;
 
+	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+	spa_pod_builder_get_state(&b.b, &state);
+
 	result.id = SPA_PARAM_EnumFormat;
 	result.next = start;
 
@@ -567,16 +551,19 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 		if (filter) {
 			struct v4l2_format fmt;
 
-			video_format = enum_filter_format(filter_media_type,
+			res = enum_filter_format(filter_media_type,
 					    filter_media_subtype,
 					    filter, port->fmtdesc.index);
-
-			if (video_format == SPA_VIDEO_FORMAT_UNKNOWN)
+			if (res == -ENOENT)
+				goto do_enum_fmt;
+			if (res < 0)
+				goto exit;
+			if (res == SPA_VIDEO_FORMAT_UNKNOWN)
 				goto enum_end;
 
 			info = find_format_info_by_media_type(filter_media_type,
 							      filter_media_subtype,
-							      video_format, 0);
+							      res, 0);
 			if (info == NULL)
 				goto next_fmtdesc;
 
@@ -602,6 +589,7 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 			}
 
 		} else {
+do_enum_fmt:
 			if ((res = xioctl(dev->fd, VIDIOC_ENUM_FMT, &port->fmtdesc)) < 0) {
 				if (errno == EINVAL)
 					goto enum_end;
@@ -632,7 +620,7 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 				goto do_frmsize;
 
 			val = spa_pod_get_values(&p->value, &n_vals, &choice);
-			if (val->type != SPA_TYPE_Rectangle)
+			if (val->type != SPA_TYPE_Rectangle || n_vals == 0)
 				goto enum_end;
 
 			if (choice == SPA_CHOICE_None) {
@@ -649,14 +637,31 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 		}
 	      do_frmsize:
 		if ((res = xioctl(dev->fd, VIDIOC_ENUM_FRAMESIZES, &port->frmsize)) < 0) {
-			if (errno == EINVAL)
+			if (errno == ENOTTY)
 				goto next_fmtdesc;
+			if (errno == EINVAL) {
+				if (port->frmsize.index == 0) {
+					port->frmsize.type = V4L2_FRMSIZE_TYPE_CONTINUOUS;
+					port->frmsize.stepwise.min_width = 16;
+					port->frmsize.stepwise.min_height = 16;
+					port->frmsize.stepwise.max_width = 16384;
+					port->frmsize.stepwise.max_height = 16384;
+					port->frmsize.stepwise.step_width = 16;
+					port->frmsize.stepwise.step_height = 16;
+					port->fmtdesc.index++;
+					port->next_fmtdesc = true;
+					goto do_frmsize_filter;
+				}
+				else
+					goto next_fmtdesc;
+			}
 
 			res = -errno;
 			spa_log_error(this->log, "'%s' VIDIOC_ENUM_FRAMESIZES: %m",
 					this->props.device);
 			goto exit;
 		}
+do_frmsize_filter:
 		if (filter) {
 			static const struct spa_rectangle step = {1, 1};
 
@@ -670,7 +675,7 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 				goto have_size;
 
 			val = spa_pod_get_values(&p->value, &n_values, &choice);
-			if (val->type != SPA_TYPE_Rectangle)
+			if (val->type != SPA_TYPE_Rectangle || n_values == 0)
 				goto have_size;
 
 			values = SPA_POD_BODY_CONST(val);
@@ -714,34 +719,35 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 		}
 	}
 
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
-	spa_pod_builder_push_object(&b, &f[0], SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
-	spa_pod_builder_add(&b,
+	spa_pod_builder_reset(&b.b, &state);
+	spa_pod_builder_push_object(&b.b, &f[0], SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
+	spa_pod_builder_add(&b.b,
 			SPA_FORMAT_mediaType,    SPA_POD_Id(info->media_type),
 			SPA_FORMAT_mediaSubtype, SPA_POD_Id(info->media_subtype),
 			0);
 
 	if (info->media_subtype == SPA_MEDIA_SUBTYPE_raw) {
-		spa_pod_builder_prop(&b, SPA_FORMAT_VIDEO_format, 0);
-		spa_pod_builder_id(&b, info->format);
+		spa_pod_builder_prop(&b.b, SPA_FORMAT_VIDEO_format, 0);
+		spa_pod_builder_id(&b.b, info->format);
 	}
-	spa_pod_builder_prop(&b, SPA_FORMAT_VIDEO_size, 0);
+
+	spa_pod_builder_prop(&b.b, SPA_FORMAT_VIDEO_size, 0);
 	if (port->frmsize.type == V4L2_FRMSIZE_TYPE_DISCRETE) {
-		spa_pod_builder_rectangle(&b,
+		spa_pod_builder_rectangle(&b.b,
 				port->frmsize.discrete.width,
 				port->frmsize.discrete.height);
 	} else if (port->frmsize.type == V4L2_FRMSIZE_TYPE_CONTINUOUS ||
 		   port->frmsize.type == V4L2_FRMSIZE_TYPE_STEPWISE) {
-		spa_pod_builder_push_choice(&b, &f[1], SPA_CHOICE_None, 0);
-		choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b, &f[1]);
+		spa_pod_builder_push_choice(&b.b, &f[1], SPA_CHOICE_None, 0);
+		choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b.b, &f[1]);
 
-		spa_pod_builder_rectangle(&b,
+		spa_pod_builder_rectangle(&b.b,
 				port->frmsize.stepwise.min_width,
 				port->frmsize.stepwise.min_height);
-		spa_pod_builder_rectangle(&b,
+		spa_pod_builder_rectangle(&b.b,
 				port->frmsize.stepwise.min_width,
 				port->frmsize.stepwise.min_height);
-		spa_pod_builder_rectangle(&b,
+		spa_pod_builder_rectangle(&b.b,
 				port->frmsize.stepwise.max_width,
 				port->frmsize.stepwise.max_height);
 
@@ -749,35 +755,43 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 			choice->body.type = SPA_CHOICE_Range;
 		} else {
 			choice->body.type = SPA_CHOICE_Step;
-			spa_pod_builder_rectangle(&b,
+			spa_pod_builder_rectangle(&b.b,
 					port->frmsize.stepwise.max_width,
 					port->frmsize.stepwise.max_height);
 		}
-		spa_pod_builder_pop(&b, &f[1]);
+		spa_pod_builder_pop(&b.b, &f[1]);
 	}
 
-	spa_pod_builder_prop(&b, SPA_FORMAT_VIDEO_framerate, 0);
+	spa_pod_builder_prop(&b.b, SPA_FORMAT_VIDEO_framerate, 0);
 
 	n_fractions = 0;
 
-	spa_pod_builder_push_choice(&b, &f[1], SPA_CHOICE_None, 0);
-	choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b, &f[1]);
+	spa_pod_builder_push_choice(&b.b, &f[1], SPA_CHOICE_None, 0);
+	choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b.b, &f[1]);
 	port->frmival.index = 0;
 
 	while (true) {
 		if ((res = xioctl(dev->fd, VIDIOC_ENUM_FRAMEINTERVALS, &port->frmival)) < 0) {
 			res = -errno;
-			if (errno == EINVAL) {
-				port->frmsize.index++;
-				port->next_frmsize = true;
-				if (port->frmival.index == 0)
-					goto next_frmsize;
-				break;
+			port->frmsize.index++;
+			port->next_frmsize = true;
+			if (errno == EINVAL || errno == ENOTTY) {
+				if (port->frmival.index == 0) {
+					port->frmival.type = V4L2_FRMIVAL_TYPE_CONTINUOUS;
+					port->frmival.stepwise.min.denominator = 1;
+					port->frmival.stepwise.min.numerator = 1;
+					port->frmival.stepwise.max.denominator = 120;
+					port->frmival.stepwise.max.numerator = 1;
+					goto do_frminterval_filter;
+				}
+				else
+					break;
 			}
 			spa_log_error(this->log, "'%s' VIDIOC_ENUM_FRAMEINTERVALS: %m",
 					this->props.device);
 			goto exit;
 		}
+do_frminterval_filter:
 		if (filter) {
 			static const struct spa_fraction step = {1, 1};
 
@@ -790,8 +804,7 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 				goto have_framerate;
 
 			val = spa_pod_get_values(&p->value, &n_values, &choice);
-
-			if (val->type != SPA_TYPE_Fraction)
+			if (val->type != SPA_TYPE_Fraction || n_values == 0)
 				goto enum_end;
 
 			values = SPA_POD_BODY(val);
@@ -830,44 +843,48 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 		if (port->frmival.type == V4L2_FRMIVAL_TYPE_DISCRETE) {
 			choice->body.type = SPA_CHOICE_Enum;
 			if (n_fractions == 0)
-				spa_pod_builder_fraction(&b,
+				spa_pod_builder_fraction(&b.b,
 							 port->frmival.discrete.denominator,
 							 port->frmival.discrete.numerator);
-			spa_pod_builder_fraction(&b,
+			spa_pod_builder_fraction(&b.b,
 						 port->frmival.discrete.denominator,
 						 port->frmival.discrete.numerator);
 			port->frmival.index++;
+			n_fractions++;
 		} else if (port->frmival.type == V4L2_FRMIVAL_TYPE_CONTINUOUS ||
 			   port->frmival.type == V4L2_FRMIVAL_TYPE_STEPWISE) {
 			if (n_fractions == 0)
-				spa_pod_builder_fraction(&b, 25, 1);
-			spa_pod_builder_fraction(&b,
+				spa_pod_builder_fraction(&b.b, 25, 1);
+			spa_pod_builder_fraction(&b.b,
 						 port->frmival.stepwise.min.denominator,
 						 port->frmival.stepwise.min.numerator);
-			spa_pod_builder_fraction(&b,
+			spa_pod_builder_fraction(&b.b,
 						 port->frmival.stepwise.max.denominator,
 						 port->frmival.stepwise.max.numerator);
 
 			if (port->frmival.type == V4L2_FRMIVAL_TYPE_CONTINUOUS) {
 				choice->body.type = SPA_CHOICE_Range;
+				n_fractions += 2;
 			} else {
 				choice->body.type = SPA_CHOICE_Step;
-				spa_pod_builder_fraction(&b,
+				spa_pod_builder_fraction(&b.b,
 							 port->frmival.stepwise.step.denominator,
 							 port->frmival.stepwise.step.numerator);
+				n_fractions += 3;
 			}
 
 			port->frmsize.index++;
 			port->next_frmsize = true;
 			break;
 		}
-		n_fractions++;
 	}
-	if (n_fractions <= 1)
+	if (n_fractions == 0)
+		goto next_frmsize;
+	if (n_fractions == 1)
 		choice->body.type = SPA_CHOICE_None;
+	spa_pod_builder_pop(&b.b, &f[1]);
 
-	spa_pod_builder_pop(&b, &f[1]);
-	result.param = spa_pod_builder_pop(&b, &f[0]);
+	result.param = spa_pod_builder_pop(&b.b, &f[0]);
 
 	spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
 
@@ -881,6 +898,49 @@ spa_v4l2_enum_format(struct impl *this, int seq,
 	return res;
 }
 
+static int probe_expbuf(struct impl *this)
+{
+	struct port *port = &this->out_ports[0];
+	struct spa_v4l2_device *dev = &port->dev;
+	struct v4l2_requestbuffers reqbuf;
+	struct v4l2_exportbuffer expbuf;
+
+	if (port->probed_expbuf)
+		return 0;
+	port->probed_expbuf = true;
+
+	spa_zero(reqbuf);
+	reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+	reqbuf.memory = V4L2_MEMORY_MMAP;
+	reqbuf.count = port->max_buffers = MAX_BUFFERS;
+
+	if (xioctl(dev->fd, VIDIOC_REQBUFS, &reqbuf) < 0) {
+		spa_log_error(this->log, "'%s' VIDIOC_REQBUFS: %m", this->props.device);
+		return -errno;
+	}
+	port->max_buffers = reqbuf.count;
+
+	spa_zero(expbuf);
+	expbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+	expbuf.index = 0;
+	expbuf.flags = O_CLOEXEC | O_RDONLY;
+	if (xioctl(dev->fd, VIDIOC_EXPBUF, &expbuf) < 0) {
+		spa_log_info(this->log, "'%s' EXPBUF not supported: %m", this->props.device);
+		port->have_expbuf = false;
+		port->alloc_buffers = false;
+	} else {
+		port->have_expbuf = true;
+		port->alloc_buffers = true;
+	}
+
+	reqbuf.count = 0;
+	if (xioctl(dev->fd, VIDIOC_REQBUFS, &reqbuf) < 0) {
+		spa_log_error(this->log, "'%s' VIDIOC_REQBUFS: %m", this->props.device);
+		return -errno;
+	}
+	return 0;
+}
+
 static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format, uint32_t flags)
 {
 	struct port *port = &this->out_ports[0];
@@ -895,8 +955,9 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 	bool match;
 
 	spa_zero(fmt);
-	spa_zero(streamparm);
 	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+
+	spa_zero(streamparm);
 	streamparm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
 
 	switch (format->media_subtype) {
@@ -924,12 +985,12 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 	info = find_format_info_by_media_type(format->media_type,
 					      format->media_subtype, video_format, 0);
 	if (info == NULL || size == NULL || framerate == NULL) {
-		spa_log_error(this->log, "unknown media type %d %d %d", format->media_type,
-			      format->media_subtype, video_format);
+		spa_log_error(this->log, "%s: unknown media type %d %d %d",
+				this->props.device, format->media_type,
+				format->media_subtype, video_format);
 		return -EINVAL;
 	}
 
-
 	fmt.fmt.pix.pixelformat = info->fourcc;
 	fmt.fmt.pix.field = V4L2_FIELD_ANY;
 	fmt.fmt.pix.width = size->width;
@@ -937,7 +998,7 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 	streamparm.parm.capture.timeperframe.numerator = framerate->denom;
 	streamparm.parm.capture.timeperframe.denominator = framerate->num;
 
-	spa_log_debug(this->log, "set %.4s %dx%d %d/%d", (char *)&fmt.fmt.pix.pixelformat,
+	spa_log_info(this->log, "set %.4s %dx%d %d/%d", (char *)&fmt.fmt.pix.pixelformat,
 		     fmt.fmt.pix.width, fmt.fmt.pix.height,
 		     streamparm.parm.capture.timeperframe.denominator,
 		     streamparm.parm.capture.timeperframe.numerator);
@@ -957,14 +1018,15 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 
 	/* some cheap USB cam's won't accept any change */
 	if (xioctl(dev->fd, VIDIOC_S_PARM, &streamparm) < 0)
-		spa_log_warn(this->log, "VIDIOC_S_PARM: %m");
+		spa_log_warn(this->log, "%s: VIDIOC_S_PARM: %m", this->props.device);
 
 	match = (reqfmt.fmt.pix.pixelformat == fmt.fmt.pix.pixelformat &&
 			reqfmt.fmt.pix.width == fmt.fmt.pix.width &&
 			reqfmt.fmt.pix.height == fmt.fmt.pix.height);
 
 	if (!match && !SPA_FLAG_IS_SET(flags, SPA_NODE_PARAM_FLAG_NEAREST)) {
-		spa_log_error(this->log, "wanted %.4s %dx%d, got %.4s %dx%d",
+		spa_log_error(this->log, "%s: wanted %.4s %dx%d, got %.4s %dx%d",
+				this->props.device,
 				(char *)&reqfmt.fmt.pix.pixelformat,
 				reqfmt.fmt.pix.width, reqfmt.fmt.pix.height,
 				(char *)&fmt.fmt.pix.pixelformat,
@@ -975,8 +1037,11 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 	if (flags & SPA_NODE_PARAM_FLAG_TEST_ONLY)
 		return match ? 0 : 1;
 
+	if (streamparm.parm.capture.timeperframe.denominator == 0)
+		streamparm.parm.capture.timeperframe.denominator = 1;
+
 	spa_log_info(this->log, "'%s' got %.4s %dx%d %d/%d",
-			this->props.device, (char *)&fmt.fmt.pix.pixelformat,
+			dev->path, (char *)&fmt.fmt.pix.pixelformat,
 			fmt.fmt.pix.width, fmt.fmt.pix.height,
 			streamparm.parm.capture.timeperframe.denominator,
 			streamparm.parm.capture.timeperframe.numerator);
@@ -984,8 +1049,8 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 	dev->have_format = true;
 	size->width = fmt.fmt.pix.width;
 	size->height = fmt.fmt.pix.height;
-	port->rate.denom = framerate->num = streamparm.parm.capture.timeperframe.denominator;
-	port->rate.num = framerate->denom = streamparm.parm.capture.timeperframe.numerator;
+
+	probe_expbuf(this);
 
 	port->fmt = fmt;
 	port->info.change_mask |= SPA_PORT_CHANGE_MASK_FLAGS | SPA_PORT_CHANGE_MASK_RATE;
@@ -993,7 +1058,8 @@ static int spa_v4l2_set_format(struct impl *this, struct spa_video_info *format,
 		SPA_PORT_FLAG_LIVE |
 		SPA_PORT_FLAG_PHYSICAL |
 		SPA_PORT_FLAG_TERMINAL;
-	port->info.rate = SPA_FRACTION(port->rate.num, port->rate.denom);
+	port->info.rate.num = streamparm.parm.capture.timeperframe.numerator;
+	port->info.rate.denom = streamparm.parm.capture.timeperframe.denominator;
 
 	return match ? 0 : 1;
 }
@@ -1046,28 +1112,38 @@ static int query_ext_ctrl_ioctl(struct port *port, struct v4l2_query_ext_ctrl *q
 	return res;
 }
 
+static struct {
+	uint32_t v4l2_id;
+	uint32_t spa_id;
+} control_map[] = {
+	{ V4L2_CID_BRIGHTNESS, SPA_PROP_brightness },
+	{ V4L2_CID_CONTRAST, SPA_PROP_contrast },
+	{ V4L2_CID_SATURATION, SPA_PROP_saturation },
+	{ V4L2_CID_HUE, SPA_PROP_hue },
+	{ V4L2_CID_GAMMA, SPA_PROP_gamma },
+	{ V4L2_CID_EXPOSURE, SPA_PROP_exposure },
+	{ V4L2_CID_GAIN, SPA_PROP_gain },
+	{ V4L2_CID_SHARPNESS, SPA_PROP_sharpness },
+};
+
 static uint32_t control_to_prop_id(struct impl *impl, uint32_t control_id)
 {
-	switch (control_id) {
-	case V4L2_CID_BRIGHTNESS:
-		return SPA_PROP_brightness;
-	case V4L2_CID_CONTRAST:
-		return SPA_PROP_contrast;
-	case V4L2_CID_SATURATION:
-		return SPA_PROP_saturation;
-	case V4L2_CID_HUE:
-		return SPA_PROP_hue;
-	case V4L2_CID_GAMMA:
-		return SPA_PROP_gamma;
-	case V4L2_CID_EXPOSURE:
-		return SPA_PROP_exposure;
-	case V4L2_CID_GAIN:
-		return SPA_PROP_gain;
-	case V4L2_CID_SHARPNESS:
-		return SPA_PROP_sharpness;
-	default:
-		return SPA_PROP_START_CUSTOM + control_id;
+	SPA_FOR_EACH_ELEMENT_VAR(control_map, c) {
+		if (c->v4l2_id == control_id)
+			return c->spa_id;
 	}
+	return SPA_PROP_START_CUSTOM + control_id;
+}
+
+static uint32_t prop_id_to_control(struct impl *impl, uint32_t prop_id)
+{
+	SPA_FOR_EACH_ELEMENT_VAR(control_map, c) {
+		if (c->spa_id == prop_id)
+			return c->v4l2_id;
+	}
+	if (prop_id >= SPA_PROP_START_CUSTOM)
+		return prop_id - SPA_PROP_START_CUSTOM;
+	return SPA_ID_INVALID;
 }
 
 static int
@@ -1079,7 +1155,8 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 	struct spa_v4l2_device *dev = &port->dev;
 	struct v4l2_query_ext_ctrl queryctrl;
 	struct spa_pod *param;
-	struct spa_pod_builder b = { 0 };
+	spa_auto(spa_pod_dynamic_builder) b = { 0 };
+	struct spa_pod_builder_state state;
 	uint32_t prop_id, ctrl_id;
 	uint8_t buffer[1024];
 	int res;
@@ -1091,6 +1168,9 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 	if ((res = spa_v4l2_open(dev, this->props.device)) < 0)
 		return res;
 
+	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+	spa_pod_builder_get_state(&b.b, &state);
+
 	result.id = SPA_PARAM_PropInfo;
 	result.next = start;
       next:
@@ -1107,6 +1187,8 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 	spa_log_debug(this->log, "test control %08x", queryctrl.id);
 
 	if (query_ext_ctrl_ioctl(port, &queryctrl) != 0) {
+		if (errno == ENOTTY)
+			goto enum_end;
 		if (errno == EINVAL) {
 			if (queryctrl.id != next_fl)
 				goto enum_end;
@@ -1123,6 +1205,7 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 		}
 		res = -errno;
 		spa_log_error(this->log, "'%s' VIDIOC_QUERYCTRL: %m", this->props.device);
+		spa_v4l2_close(dev);
 		return res;
 	}
 	if (result.next & next_fl)
@@ -1138,7 +1221,7 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 
 	ctrl_id = queryctrl.id & ~next_fl;
 
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+	spa_pod_builder_reset(&b.b, &state);
 
 	prop_id = control_to_prop_id(this, ctrl_id);
 
@@ -1148,61 +1231,55 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 
 	spa_log_debug(this->log, "Control '%s' %d %d", queryctrl.name, prop_id, ctrl_id);
 
-	port->n_controls++;
-
 	switch (queryctrl.type) {
 	case V4L2_CTRL_TYPE_INTEGER:
-		param = spa_pod_builder_add_object(&b,
+		port->controls[port->n_controls].type = SPA_TYPE_Int;
+		param = spa_pod_builder_add_object(&b.b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_id,   SPA_POD_Id(prop_id),
 			SPA_PROP_INFO_type, SPA_POD_CHOICE_STEP_Int(
-							queryctrl.default_value,
-							queryctrl.minimum,
-							queryctrl.maximum,
-							queryctrl.step),
-			SPA_PROP_INFO_name, SPA_POD_String(queryctrl.name));
+							(int32_t)queryctrl.default_value,
+							(int32_t)queryctrl.minimum,
+							(int32_t)queryctrl.maximum,
+							(int32_t)queryctrl.step),
+			SPA_PROP_INFO_description, SPA_POD_String(queryctrl.name));
 		break;
 	case V4L2_CTRL_TYPE_BOOLEAN:
-		param = spa_pod_builder_add_object(&b,
+		port->controls[port->n_controls].type = SPA_TYPE_Bool;
+		param = spa_pod_builder_add_object(&b.b,
 			SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo,
 			SPA_PROP_INFO_id,   SPA_POD_Id(prop_id),
-			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool(queryctrl.default_value),
-			SPA_PROP_INFO_name, SPA_POD_String(queryctrl.name));
+			SPA_PROP_INFO_type, SPA_POD_CHOICE_Bool((bool)queryctrl.default_value),
+			SPA_PROP_INFO_description, SPA_POD_String(queryctrl.name));
 		break;
 	case V4L2_CTRL_TYPE_MENU:
 	{
 		struct v4l2_querymenu querymenu;
-		struct spa_pod_builder_state state;
 
-		spa_pod_builder_push_object(&b, &f[0], SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo);
-		spa_pod_builder_add(&b,
+		port->controls[port->n_controls].type = SPA_TYPE_Int;
+		spa_pod_builder_push_object(&b.b, &f[0], SPA_TYPE_OBJECT_PropInfo, SPA_PARAM_PropInfo);
+		spa_pod_builder_add(&b.b,
 			SPA_PROP_INFO_id,    SPA_POD_Id(prop_id),
-			SPA_PROP_INFO_type,  SPA_POD_CHOICE_ENUM_Int(1, queryctrl.default_value),
-			SPA_PROP_INFO_name,  SPA_POD_String(queryctrl.name),
+			SPA_PROP_INFO_type,  SPA_POD_CHOICE_ENUM_Int(1, (int32_t)queryctrl.default_value),
+			SPA_PROP_INFO_description,  SPA_POD_String(queryctrl.name),
 			0);
 
 		spa_zero(querymenu);
 		querymenu.id = queryctrl.id;
 
-		spa_pod_builder_prop(&b, SPA_PROP_INFO_labels, 0);
+		spa_pod_builder_prop(&b.b, SPA_PROP_INFO_labels, 0);
 
-		spa_pod_builder_get_state(&b, &state);
-		spa_pod_builder_push_struct(&b, &f[1]);
+		spa_pod_builder_push_struct(&b.b, &f[1]);
 		for (querymenu.index = queryctrl.minimum;
 		    querymenu.index <= queryctrl.maximum;
 		    querymenu.index++) {
 			if (xioctl(dev->fd, VIDIOC_QUERYMENU, &querymenu) == 0) {
-				spa_pod_builder_int(&b, querymenu.index);
-				spa_pod_builder_string(&b, (const char *)querymenu.name);
+				spa_pod_builder_int(&b.b, querymenu.index);
+				spa_pod_builder_string(&b.b, (const char *)querymenu.name);
 			}
 		}
-		if (spa_pod_builder_pop(&b, &f[1]) == NULL) {
-			spa_log_warn(this->log, "can't create Control '%s' overflow %d",
-					queryctrl.name, b.state.offset);
-			spa_pod_builder_reset(&b, &state);
-			spa_pod_builder_none(&b);
-		}
-		param = spa_pod_builder_pop(&b, &f[0]);
+		spa_pod_builder_pop(&b.b, &f[1]);
+		param = spa_pod_builder_pop(&b.b, &f[0]);
 		break;
 	}
 	case V4L2_CTRL_TYPE_INTEGER_MENU:
@@ -1214,7 +1291,10 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 		goto next;
 
 	}
-	if (spa_pod_filter(&b, &result.param, param, filter) < 0)
+
+	port->n_controls++;
+
+	if (spa_pod_filter(&b.b, &result.param, param, filter) < 0)
 		goto next;
 
 	spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
@@ -1228,6 +1308,90 @@ spa_v4l2_enum_controls(struct impl *this, int seq,
 	return res;
 }
 
+static int
+spa_v4l2_update_controls(struct impl *this)
+{
+	struct port *port = &this->out_ports[0];
+	struct spa_v4l2_device *dev = &port->dev;
+	int res;
+	uint32_t i;
+
+	if ((res = spa_v4l2_open(dev, this->props.device)) < 0)
+		return res;
+
+	for (i = 0; i < port->n_controls; i++) {
+		struct control *c = &port->controls[i];
+		struct v4l2_control control;
+
+		spa_zero(control);
+		control.id = c->ctrl_id;
+		if (xioctl(dev->fd, VIDIOC_G_CTRL, &control) < 0) {
+			/* Write only controls like relative pan/tilt return EACCES */
+			if (errno == EACCES) {
+				c->value = 0;
+				continue;
+			}
+			res = -errno;
+			goto done;
+		}
+		c->value = control.value;
+	}
+	res = 0;
+done:
+	spa_v4l2_close(dev);
+	return res;
+}
+
+static int
+spa_v4l2_set_control(struct impl *this, uint32_t id,
+		       const struct spa_pod_prop *prop)
+{
+	struct port *port = &this->out_ports[0];
+	struct spa_v4l2_device *dev = &port->dev;
+	struct v4l2_control control;
+	int res;
+
+	spa_zero(control);
+	control.id = prop_id_to_control(this, prop->key);
+	if (control.id == SPA_ID_INVALID)
+		return -ENOENT;
+
+	if ((res = spa_v4l2_open(dev, this->props.device)) < 0)
+		return res;
+
+	switch (SPA_POD_TYPE(&prop->value)) {
+	case SPA_TYPE_Bool:
+	{
+		bool val;
+		if ((res = spa_pod_get_bool(&prop->value, &val)) < 0)
+			goto done;
+		control.value = val;
+		break;
+	}
+	case SPA_TYPE_Int:
+	{
+		int32_t val;
+		if ((res = spa_pod_get_int(&prop->value, &val)) < 0)
+			goto done;
+		control.value = val;
+		break;
+	}
+	default:
+		res = -EINVAL;
+		goto done;
+	}
+	if (xioctl(dev->fd, VIDIOC_S_CTRL, &control) < 0) {
+		res = -errno;
+		goto done;
+	}
+
+	res = 0;
+
+done:
+	spa_v4l2_close(dev);
+	return res;
+}
+
 static int mmap_read(struct impl *this)
 {
 	struct port *port = &this->out_ports[0];
@@ -1244,17 +1408,32 @@ static int mmap_read(struct impl *this)
 	if (xioctl(dev->fd, VIDIOC_DQBUF, &buf) < 0)
 		return -errno;
 
+	spa_log_trace(this->log, "v4l2 %p: have output %d/%d", this, buf.index, buf.sequence);
+
+	/* Drop the first frame in order to work around common firmware
+	 * timestamp issues */
+	if (port->first_buffer) {
+		port->first_buffer = false;
+		if (xioctl(dev->fd, VIDIOC_QBUF, &buf) < 0)
+			spa_log_warn(this->log, "v4l2 %p: error qbuf: %m", this);
+		return 0;
+	}
+
 	pts = SPA_TIMEVAL_TO_NSEC(&buf.timestamp);
-	spa_log_trace(this->log, "v4l2 %p: have output %d", this, buf.index);
 
 	if (this->clock) {
+		/* FIXME, we should follow the driver clock and target_ values.
+		 * for now we ignore and use our own. */
+		this->clock->target_rate = port->info.rate;
+		this->clock->target_duration = 1;
+
 		this->clock->nsec = pts;
-		this->clock->rate = port->rate;
+		this->clock->rate = port->info.rate;
 		this->clock->position = buf.sequence;
 		this->clock->duration = 1;
 		this->clock->delay = 0;
 		this->clock->rate_diff = 1.0;
-		this->clock->next_nsec = pts + 1000000000LL / port->rate.denom;
+		this->clock->next_nsec = pts + port->info.rate.num * SPA_NSEC_PER_SEC / port->info.rate.denom;
 	}
 
 	b = &port->buffers[buf.index];
@@ -1267,15 +1446,21 @@ static int mmap_read(struct impl *this)
 		b->h->pts = pts;
 		b->h->dts_offset = 0;
 	}
+	if (b->vt) {
+		b->vt->transform = this->transform;
+	}
 
 	d = b->outbuf->datas;
 	d[0].chunk->offset = 0;
-	d[0].chunk->size = buf.bytesused;
+	d[0].chunk->size = SPA_MIN(buf.bytesused, d[0].maxsize);
 	d[0].chunk->stride = port->fmt.fmt.pix.bytesperline;
 	d[0].chunk->flags = 0;
 	if (buf.flags & V4L2_BUF_FLAG_ERROR)
 		d[0].chunk->flags |= SPA_CHUNK_FLAG_CORRUPTED;
 
+	if (b->mmap_ptr && b->ptr)
+		memcpy(b->ptr, b->mmap_ptr, d[0].chunk->size);
+
 	spa_list_append(&port->queue, &b->link);
 	return 0;
 }
@@ -1286,6 +1471,7 @@ static void v4l2_on_fd_events(struct spa_source *source)
 	struct spa_io_buffers *io;
 	struct port *port = &this->out_ports[0];
 	struct buffer *b;
+	int res;
 
 	if (source->rmask & SPA_IO_ERR) {
 		struct port *port = &this->out_ports[0];
@@ -1300,14 +1486,22 @@ static void v4l2_on_fd_events(struct spa_source *source)
 		return;
 	}
 
-	if (mmap_read(this) < 0)
+	if ((res = mmap_read(this)) < 0) {
+		spa_log_warn(this->log, "v4l2 %p: mmap read error:%s", this, spa_strerror(res));
 		return;
+	}
 
 	if (spa_list_is_empty(&port->queue))
 		return;
 
 	io = port->io;
-	if (io != NULL && io->status != SPA_STATUS_HAVE_DATA) {
+	if (io == NULL) {
+		b = spa_list_first(&port->queue, struct buffer, link);
+		spa_list_remove(&b->link);
+		SPA_FLAG_SET(b->flags, BUFFER_FLAG_OUTSTANDING);
+		spa_v4l2_buffer_recycle(this, b->id);
+	}
+	else if (io->status != SPA_STATUS_HAVE_DATA) {
 		if (io->buffer_id < port->n_buffers)
 			spa_v4l2_buffer_recycle(this, io->buffer_id);
 
@@ -1339,7 +1533,8 @@ static int spa_v4l2_use_buffers(struct impl *this, struct spa_buffer **buffers,
 		} else if (d[0].type == SPA_DATA_DmaBuf) {
 			port->memtype = V4L2_MEMORY_DMABUF;
 		} else {
-			spa_log_error(this->log, "can't use buffers of type %d", d[0].type);
+			spa_log_error(this->log, "%s: can't use buffers of type %d",
+					this->props.device, d[0].type);
 			return -EINVAL;
 		}
 	}
@@ -1350,8 +1545,22 @@ static int spa_v4l2_use_buffers(struct impl *this, struct spa_buffer **buffers,
 	reqbuf.count = n_buffers;
 
 	if (xioctl(dev->fd, VIDIOC_REQBUFS, &reqbuf) < 0) {
-		spa_log_error(this->log, "'%s' VIDIOC_REQBUFS %m", this->props.device);
-		return -errno;
+		if (port->memtype != V4L2_MEMORY_USERPTR) {
+			spa_log_error(this->log, "'%s' VIDIOC_REQBUFS %m", this->props.device);
+			return -errno;
+		}
+		/* some drivers (v4l2loopback) don't support USERPTR
+		 * and so we need to try again with MMAP and memcpy */
+		port->memtype = V4L2_MEMORY_MMAP;
+		spa_zero(reqbuf);
+		reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+		reqbuf.memory = port->memtype;
+		reqbuf.count = n_buffers;
+
+		if (xioctl(dev->fd, VIDIOC_REQBUFS, &reqbuf) < 0) {
+			spa_log_error(this->log, "'%s' VIDIOC_REQBUFS %m", this->props.device);
+			return -errno;
+		}
 	}
 	spa_log_debug(this->log, "got %d buffers", reqbuf.count);
 	if (reqbuf.count < n_buffers) {
@@ -1368,11 +1577,13 @@ static int spa_v4l2_use_buffers(struct impl *this, struct spa_buffer **buffers,
 		b->outbuf = buffers[i];
 		b->flags = BUFFER_FLAG_OUTSTANDING;
 		b->h = spa_buffer_find_meta_data(buffers[i], SPA_META_Header, sizeof(*b->h));
+		b->vt = spa_buffer_find_meta_data(buffers[i], SPA_META_VideoTransform, sizeof(*b->vt));
 
-		spa_log_debug(this->log, "import buffer %p", buffers[i]);
+		spa_log_debug(this->log, "%s: import buffer %p", this->props.device, buffers[i]);
 
 		if (buffers[i]->n_datas < 1) {
-			spa_log_error(this->log, "invalid memory on buffer %p", buffers[i]);
+			spa_log_error(this->log, "%s: invalid memory on buffer %p",
+					this->props.device, buffers[i]);
 			return -EINVAL;
 		}
 		d = buffers[i]->datas;
@@ -1382,7 +1593,8 @@ static int spa_v4l2_use_buffers(struct impl *this, struct spa_buffer **buffers,
 		b->v4l2_buffer.memory = port->memtype;
 		b->v4l2_buffer.index = i;
 
-		if (port->memtype == V4L2_MEMORY_USERPTR) {
+		if (port->memtype == V4L2_MEMORY_USERPTR ||
+		    port->memtype == V4L2_MEMORY_MMAP) {
 			if (d[0].data == NULL) {
 				void *data;
 
@@ -1400,15 +1612,31 @@ static int spa_v4l2_use_buffers(struct impl *this, struct spa_buffer **buffers,
 			else
 				b->ptr = d[0].data;
 
-			b->v4l2_buffer.m.userptr = (unsigned long) b->ptr;
-			b->v4l2_buffer.length = d[0].maxsize;
+			if (port->memtype == V4L2_MEMORY_USERPTR) {
+				b->v4l2_buffer.m.userptr = (unsigned long) b->ptr;
+				b->v4l2_buffer.length = d[0].maxsize;
+			}
+			else {
+				if (xioctl(dev->fd, VIDIOC_QUERYBUF, &b->v4l2_buffer) < 0) {
+					spa_log_error(this->log, "'%s' VIDIOC_QUERYBUF: %m", this->props.device);
+					return -errno;
+				}
+				b->mmap_ptr = mmap(NULL,
+						b->v4l2_buffer.length,
+						PROT_READ, MAP_PRIVATE,
+						dev->fd, b->v4l2_buffer.m.offset);
+				if (b->mmap_ptr == MAP_FAILED) {
+					spa_log_error(this->log, "'%s' mmap: %m", this->props.device);
+					return -errno;
+				}
+			}
 		}
 		else if (port->memtype == V4L2_MEMORY_DMABUF) {
 			b->v4l2_buffer.m.fd = d[0].fd;
 		}
 		else {
-			spa_log_error(this->log, "invalid port memory %d",
-					port->memtype);
+			spa_log_error(this->log, "%s: invalid port memory %d",
+					this->props.device, port->memtype);
 			return -EIO;
 		}
 
@@ -1442,11 +1670,10 @@ mmap_init(struct impl *this,
 	}
 
 	spa_log_debug(this->log, "got %d buffers", reqbuf.count);
-	n_buffers = reqbuf.count;
 
-	if (n_buffers < 2) {
-		spa_log_error(this->log, "'%s' can't allocate enough buffers (%d)",
-				this->props.device, n_buffers);
+	if (reqbuf.count < n_buffers) {
+		spa_log_error(this->log, "'%s' can't allocate enough buffers (%d < %d)",
+				this->props.device, reqbuf.count, n_buffers);
 		return -ENOMEM;
 	}
 
@@ -1455,7 +1682,8 @@ mmap_init(struct impl *this,
 		struct spa_data *d;
 
 		if (buffers[i]->n_datas < 1) {
-			spa_log_error(this->log, "invalid buffer data");
+			spa_log_error(this->log, "%s: invalid buffer data",
+					this->props.device);
 			return -EINVAL;
 		}
 
@@ -1464,6 +1692,7 @@ mmap_init(struct impl *this,
 		b->outbuf = buffers[i];
 		b->flags = BUFFER_FLAG_OUTSTANDING;
 		b->h = spa_buffer_find_meta_data(buffers[i], SPA_META_Header, sizeof(*b->h));
+		b->vt = spa_buffer_find_meta_data(buffers[i], SPA_META_VideoTransform, sizeof(*b->vt));
 
 		spa_zero(b->v4l2_buffer);
 		b->v4l2_buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
@@ -1492,6 +1721,7 @@ mmap_init(struct impl *this,
 
 		spa_log_debug(this->log, "data types %08x", d[0].type);
 
+again:
 		if (port->have_expbuf &&
 		    d[0].type != SPA_ID_INVALID &&
 		    (d[0].type & ((1u << SPA_DATA_DmaBuf)|(1u<<SPA_DATA_MemFd)))) {
@@ -1506,7 +1736,7 @@ mmap_init(struct impl *this,
 					spa_log_debug(this->log, "'%s' VIDIOC_EXPBUF not supported: %m",
 							this->props.device);
 					port->have_expbuf = false;
-					goto fallback;
+					goto again;
 				}
 				spa_log_error(this->log, "'%s' VIDIOC_EXPBUF: %m", this->props.device);
 				return -errno;
@@ -1515,14 +1745,13 @@ mmap_init(struct impl *this,
 				d[0].type = SPA_DATA_DmaBuf;
 			else
 				d[0].type = SPA_DATA_MemFd;
-			d[0].flags = SPA_DATA_FLAG_READABLE;
+			d[0].flags = SPA_DATA_FLAG_READABLE | SPA_DATA_FLAG_MAPPABLE;
 			d[0].fd = expbuf.fd;
 			d[0].data = NULL;
 			SPA_FLAG_SET(b->flags, BUFFER_FLAG_ALLOCATED);
 			spa_log_debug(this->log, "EXPBUF fd:%d", expbuf.fd);
 			use_expbuf = true;
 		} else if (d[0].type & (1u << SPA_DATA_MemPtr)) {
-fallback:
 			d[0].type = SPA_DATA_MemPtr;
 			d[0].flags = SPA_DATA_FLAG_READABLE;
 			d[0].fd = -1;
@@ -1541,12 +1770,14 @@ fallback:
 			spa_log_debug(this->log, "mmap offset:%u data:%p", d[0].mapoffset, b->ptr);
 			use_expbuf = false;
 		} else {
-			spa_log_error(this->log, "unsupported data type:%08x", d[0].type);
+			spa_log_error(this->log, "%s: unsupported data type:%08x",
+					this->props.device, d[0].type);
+			port->alloc_buffers = false;
 			return -ENOTSUP;
 		}
 		spa_v4l2_buffer_recycle(this, i);
 	}
-	spa_log_info(this->log, "have %u buffers using %s", n_buffers,
+	spa_log_info(this->log, "%s: have %u buffers using %s", dev->path, n_buffers,
 			use_expbuf ? "EXPBUF" : "MMAP");
 
 	port->n_buffers = n_buffers;
@@ -1584,8 +1815,8 @@ spa_v4l2_alloc_buffers(struct impl *this,
 		if ((res = read_init(this)) < 0)
 			return res;
 	} else {
-		spa_log_error(this->log, "invalid capabilities %08x",
-					dev->cap.capabilities);
+		spa_log_error(this->log, "%s: invalid capabilities %08x",
+				this->props.device, dev->cap.capabilities);
 		return -EIO;
 	}
 
@@ -1609,6 +1840,9 @@ static int spa_v4l2_stream_on(struct impl *this)
 
 	spa_log_debug(this->log, "starting");
 
+	port->first_buffer = true;
+	mmap_read(this);
+
 	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
 	if (xioctl(dev->fd, VIDIOC_STREAMON, &type) < 0) {
 		spa_log_error(this->log, "'%s' VIDIOC_STREAMON: %m", this->props.device);
