@@ -1,26 +1,6 @@
-/* PipeWire
- *
- * Copyright © 2020 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2020 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <spa/utils/string.h>
 #include <spa/debug/types.h>
@@ -33,8 +13,8 @@
 
 static const struct format audio_formats[] = {
 	[SAMPLE_U8] = { SAMPLE_U8, SPA_AUDIO_FORMAT_U8, "u8", 1 },
-	[SAMPLE_ALAW] = { SAMPLE_ALAW, SPA_AUDIO_FORMAT_ALAW, "aLaw", 1 },
-	[SAMPLE_ULAW] = { SAMPLE_ULAW, SPA_AUDIO_FORMAT_ULAW, "uLaw", 1 },
+	[SAMPLE_ALAW] = { SAMPLE_ALAW, SPA_AUDIO_FORMAT_ALAW, "alaw", 1 },
+	[SAMPLE_ULAW] = { SAMPLE_ULAW, SPA_AUDIO_FORMAT_ULAW, "ulaw", 1 },
 	[SAMPLE_S16LE] = { SAMPLE_S16LE, SPA_AUDIO_FORMAT_S16_LE, "s16le", 2 },
 	[SAMPLE_S16BE] = { SAMPLE_S16BE, SPA_AUDIO_FORMAT_S16_BE, "s16be", 2 },
 	[SAMPLE_FLOAT32LE] = { SAMPLE_FLOAT32LE, SPA_AUDIO_FORMAT_F32_LE, "float32le", 4 },
@@ -46,19 +26,30 @@ static const struct format audio_formats[] = {
 	[SAMPLE_S24_32LE] = { SAMPLE_S24_32LE, SPA_AUDIO_FORMAT_S24_32_LE, "s24-32le", 4 },
 	[SAMPLE_S24_32BE] = { SAMPLE_S24_32BE, SPA_AUDIO_FORMAT_S24_32_BE, "s24-32be", 4 },
 
-#if __BYTE_ORDER == __BIG_ENDIAN
-	{ SAMPLE_S16BE, SPA_AUDIO_FORMAT_S16_BE, "s16ne", 2 },
-	{ SAMPLE_FLOAT32BE, SPA_AUDIO_FORMAT_F32_BE, "float32ne", 4 },
-	{ SAMPLE_S32BE, SPA_AUDIO_FORMAT_S32_BE, "s32ne", 4 },
-	{ SAMPLE_S24BE, SPA_AUDIO_FORMAT_S24_BE, "s24ne", 3 },
-	{ SAMPLE_S24_32BE, SPA_AUDIO_FORMAT_S24_32_BE, "s24-32ne", 4 },
-#elif __BYTE_ORDER == __LITTLE_ENDIAN
-	{ SAMPLE_S16LE, SPA_AUDIO_FORMAT_S16_LE, "s16ne", 2 },
-	{ SAMPLE_FLOAT32LE, SPA_AUDIO_FORMAT_F32_LE, "float32ne", 4 },
-	{ SAMPLE_S32LE, SPA_AUDIO_FORMAT_S32_LE, "s32ne", 4 },
-	{ SAMPLE_S24LE, SPA_AUDIO_FORMAT_S24_LE, "s24ne", 3 },
-	{ SAMPLE_S24_32LE, SPA_AUDIO_FORMAT_S24_32_LE, "s24-32ne", 4 },
-#endif
+	{ SAMPLE_U8, SPA_AUDIO_FORMAT_U8, "8", 1 },
+	{ SAMPLE_ULAW, SPA_AUDIO_FORMAT_ULAW, "mulaw", 1 },
+
+	{ SAMPLE_S16NE, SPA_AUDIO_FORMAT_S16, "s16ne", 2 },
+	{ SAMPLE_S16NE, SPA_AUDIO_FORMAT_S16, "s16", 2 },
+	{ SAMPLE_S16NE, SPA_AUDIO_FORMAT_S16, "16", 2 },
+	{ SAMPLE_FLOAT32NE, SPA_AUDIO_FORMAT_F32, "float32ne", 4 },
+	{ SAMPLE_FLOAT32NE, SPA_AUDIO_FORMAT_F32, "float32", 4 },
+	{ SAMPLE_FLOAT32NE, SPA_AUDIO_FORMAT_F32, "float", 4 },
+	{ SAMPLE_S32NE, SPA_AUDIO_FORMAT_S32, "s32ne", 4 },
+	{ SAMPLE_S32NE, SPA_AUDIO_FORMAT_S32, "s32", 4 },
+	{ SAMPLE_S32NE, SPA_AUDIO_FORMAT_S32, "32", 4 },
+	{ SAMPLE_S24NE, SPA_AUDIO_FORMAT_S24, "s24ne", 3 },
+	{ SAMPLE_S24NE, SPA_AUDIO_FORMAT_S24, "s24", 3 },
+	{ SAMPLE_S24NE, SPA_AUDIO_FORMAT_S24, "24", 3 },
+	{ SAMPLE_S24_32NE, SPA_AUDIO_FORMAT_S24_32, "s24-32ne", 4 },
+	{ SAMPLE_S24_32NE, SPA_AUDIO_FORMAT_S24_32, "s24-32", 4 },
+
+	{ SAMPLE_S16RE, SPA_AUDIO_FORMAT_S16_OE, "s16re", 2 },
+	{ SAMPLE_FLOAT32RE, SPA_AUDIO_FORMAT_F32_OE, "float32re", 4 },
+	{ SAMPLE_S32RE, SPA_AUDIO_FORMAT_S32_OE, "s32re", 4 },
+	{ SAMPLE_S24RE, SPA_AUDIO_FORMAT_S24_OE, "s24re", 3 },
+	{ SAMPLE_S24_32RE, SPA_AUDIO_FORMAT_S24_32_OE, "s24-32re", 4 },
+
 	/* planar formats, we just report them as interleaved */
 	{ SAMPLE_U8, SPA_AUDIO_FORMAT_U8P, "u8ne", 1 },
 	{ SAMPLE_S16NE, SPA_AUDIO_FORMAT_S16P, "s16ne", 2 },
@@ -154,37 +145,34 @@ uint32_t format_name2id(const char *name)
 		if (spa_streq(name, spa_debug_type_short_name(spa_type_audio_format[i].name)))
 			return spa_type_audio_format[i].type;
 	}
-	return SPA_AUDIO_CHANNEL_UNKNOWN;
+	return SPA_AUDIO_FORMAT_UNKNOWN;
 }
 
 uint32_t format_paname2id(const char *name, size_t size)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_formats); i++) {
-		if (audio_formats[i].name != NULL &&
-		    strncmp(name, audio_formats[i].name, size) == 0)
-			return audio_formats[i].id;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_formats, f) {
+		if (f->name != NULL &&
+		    strncasecmp(name, f->name, size) == 0 &&
+		    strlen(f->name) == size)
+			return f->id;
 	}
 	return SPA_AUDIO_FORMAT_UNKNOWN;
 }
 
 enum sample_format format_id2pa(uint32_t id)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_formats); i++) {
-		if (id == audio_formats[i].id)
-			return audio_formats[i].pa;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_formats, f) {
+		if (id == f->id)
+			return f->pa;
 	}
 	return SAMPLE_INVALID;
 }
 
 const char *format_id2paname(uint32_t id)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_formats); i++) {
-		if (id == audio_formats[i].id &&
-		    audio_formats[i].name != NULL)
-			return audio_formats[i].name;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_formats, f) {
+		if (id == f->id && f->name != NULL)
+			return f->name;
 	}
 	return "invalid";
 }
@@ -193,16 +181,31 @@ uint32_t sample_spec_frame_size(const struct sample_spec *ss)
 {
 	switch (ss->format) {
 	case SPA_AUDIO_FORMAT_U8:
+	case SPA_AUDIO_FORMAT_U8P:
+	case SPA_AUDIO_FORMAT_S8:
+	case SPA_AUDIO_FORMAT_S8P:
 	case SPA_AUDIO_FORMAT_ULAW:
 	case SPA_AUDIO_FORMAT_ALAW:
 		return ss->channels;
 	case SPA_AUDIO_FORMAT_S16_LE:
 	case SPA_AUDIO_FORMAT_S16_BE:
 	case SPA_AUDIO_FORMAT_S16P:
+	case SPA_AUDIO_FORMAT_U16_LE:
+	case SPA_AUDIO_FORMAT_U16_BE:
 		return 2 * ss->channels;
 	case SPA_AUDIO_FORMAT_S24_LE:
 	case SPA_AUDIO_FORMAT_S24_BE:
 	case SPA_AUDIO_FORMAT_S24P:
+	case SPA_AUDIO_FORMAT_U24_LE:
+	case SPA_AUDIO_FORMAT_U24_BE:
+	case SPA_AUDIO_FORMAT_S20_LE:
+	case SPA_AUDIO_FORMAT_S20_BE:
+	case SPA_AUDIO_FORMAT_U20_LE:
+	case SPA_AUDIO_FORMAT_U20_BE:
+	case SPA_AUDIO_FORMAT_S18_LE:
+	case SPA_AUDIO_FORMAT_S18_BE:
+	case SPA_AUDIO_FORMAT_U18_LE:
+	case SPA_AUDIO_FORMAT_U18_BE:
 		return 3 * ss->channels;
 	case SPA_AUDIO_FORMAT_F32_LE:
 	case SPA_AUDIO_FORMAT_F32_BE:
@@ -210,10 +213,18 @@ uint32_t sample_spec_frame_size(const struct sample_spec *ss)
 	case SPA_AUDIO_FORMAT_S32_LE:
 	case SPA_AUDIO_FORMAT_S32_BE:
 	case SPA_AUDIO_FORMAT_S32P:
+	case SPA_AUDIO_FORMAT_U32_LE:
+	case SPA_AUDIO_FORMAT_U32_BE:
 	case SPA_AUDIO_FORMAT_S24_32_LE:
 	case SPA_AUDIO_FORMAT_S24_32_BE:
 	case SPA_AUDIO_FORMAT_S24_32P:
+	case SPA_AUDIO_FORMAT_U24_32_LE:
+	case SPA_AUDIO_FORMAT_U24_32_BE:
 		return 4 * ss->channels;
+	case SPA_AUDIO_FORMAT_F64_LE:
+	case SPA_AUDIO_FORMAT_F64_BE:
+	case SPA_AUDIO_FORMAT_F64P:
+		return 8 * ss->channels;
 	default:
 		return 0;
 	}
@@ -226,6 +237,48 @@ bool sample_spec_valid(const struct sample_spec *ss)
 	    ss->channels > 0 && ss->channels <= CHANNELS_MAX);
 }
 
+void sample_spec_fix(struct sample_spec *ss, struct channel_map *map,
+		const struct sample_spec *fix_ss, const struct channel_map *fix_map,
+		struct spa_dict *props)
+{
+	const char *str;
+	if (fix_ss->format != 0) {
+		if ((str = spa_dict_lookup(props, "pulse.fix.format")) != NULL) {
+			uint32_t val = format_name2id(str);
+			if (val != SPA_AUDIO_FORMAT_UNKNOWN)
+				ss->format = val;
+		}
+		else
+			ss->format = fix_ss->format;
+		/* convert back and forth to convert potential planar to packed */
+		ss->format = format_pa2id(format_id2pa(ss->format));
+	}
+	if (fix_ss->rate != 0) {
+		if ((str = spa_dict_lookup(props, "pulse.fix.rate")) != NULL) {
+			uint32_t val = atoi(str);
+			if (val != 0)
+				ss->rate = val;
+		}
+		else
+			ss->rate = fix_ss->rate;
+		ss->rate = SPA_CLAMP(ss->rate, 0u, RATE_MAX);
+	}
+	if (fix_ss->channels != 0) {
+		if ((str = spa_dict_lookup(props, "pulse.fix.position")) != NULL) {
+			struct channel_map val;
+			channel_map_parse_position(str, &val);
+			if (val.channels > 0) {
+				ss->channels = val.channels;
+				*map = val;
+			}
+		} else {
+			ss->channels = fix_ss->channels;
+			*map = *fix_map;
+		}
+		ss->channels = SPA_CLAMP(ss->channels, 0u, CHANNELS_MAX);
+	}
+}
+
 uint32_t channel_pa2id(enum channel_position channel)
 {
 	if (channel < 0 || (size_t)channel >= SPA_N_ELEMENTS(audio_channels))
@@ -266,21 +319,29 @@ enum channel_position channel_id2pa(uint32_t id, uint32_t *aux)
 
 const char *channel_id2paname(uint32_t id, uint32_t *aux)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_channels); i++) {
-		if (id == audio_channels[i].channel &&
-		    audio_channels[i].name != NULL)
-			return audio_channels[i].name;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_channels, i) {
+		if (id == i->channel && i->name != NULL)
+			return i->name;
 	}
 	return audio_channels[CHANNEL_POSITION_AUX0 + ((*aux)++ & 31)].name;
 }
 
+#define spa_streqn(a,s,b) (strlen(b) == (s) && strncmp((a),(b),(s)) == 0)
+
 uint32_t channel_paname2id(const char *name, size_t size)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_channels); i++) {
-		if (strncmp(name, audio_channels[i].name, size) == 0)
-			return audio_channels[i].channel;
+	if (spa_streqn(name, size, "left"))
+		return SPA_AUDIO_CHANNEL_FL;
+	else if (spa_streqn(name, size, "right"))
+		return SPA_AUDIO_CHANNEL_FR;
+	else if (spa_streqn(name, size, "center"))
+		return SPA_AUDIO_CHANNEL_FC;
+	else if (spa_streqn(name, size, "subwoofer"))
+		return SPA_AUDIO_CHANNEL_LFE;
+
+	SPA_FOR_EACH_ELEMENT_VAR(audio_channels, i) {
+		if (spa_streqn(name, size, i->name))
+			return i->channel;
 	}
 	return SPA_AUDIO_CHANNEL_UNKNOWN;
 }
@@ -297,6 +358,7 @@ void channel_map_parse(const char *str, struct channel_map *map)
 {
 	const char *p = str;
 	size_t len;
+	uint32_t channels = 0;
 
 	if (spa_streq(p, "stereo")) {
 		*map = (struct channel_map) {
@@ -360,13 +422,35 @@ void channel_map_parse(const char *str, struct channel_map *map)
 			.map[7] = SPA_AUDIO_CHANNEL_SR,
 		};
 	} else {
-		map->channels = 0;
-		while (*p && map->channels < SPA_AUDIO_MAX_CHANNELS) {
+		channels = map->channels = 0;
+		while (*p && channels < SPA_AUDIO_MAX_CHANNELS) {
+			uint32_t chname;
+
 			if ((len = strcspn(p, ",")) == 0)
 				break;
-			map->map[map->channels++] = channel_paname2id(p, len);
+			chname = channel_paname2id(p, len);
+			if (chname == SPA_AUDIO_CHANNEL_UNKNOWN)
+				return;
+			map->map[channels++] = chname;
 			p += len + strspn(p+len, ",");
 		}
+		map->channels = channels;
+	}
+}
+
+void channel_map_parse_position(const char *str, struct channel_map *map)
+{
+	struct spa_json it[2];
+	char v[256];
+
+	spa_json_init(&it[0], str, strlen(str));
+	if (spa_json_enter_array(&it[0], &it[1]) <= 0)
+		spa_json_init(&it[1], str, strlen(str));
+
+	map->channels = 0;
+	while (spa_json_get_string(&it[1], v, sizeof(v)) > 0 &&
+	    map->channels < SPA_AUDIO_MAX_CHANNELS) {
+		map->map[map->channels++] = channel_name2id(v);
 	}
 }
 
@@ -424,26 +508,8 @@ static enum encoding format_encoding_from_id(uint32_t id)
 	return ENCODING_ANY;
 }
 
-static inline int
-audio_raw_parse_opt(const struct spa_pod *format, struct spa_audio_info_raw *info)
-{
-	struct spa_pod *position = NULL;
-	int res;
-	info->flags = 0;
-	res = spa_pod_parse_object(format,
-			SPA_TYPE_OBJECT_Format, NULL,
-			SPA_FORMAT_AUDIO_format,        SPA_POD_OPT_Id(&info->format),
-			SPA_FORMAT_AUDIO_rate,          SPA_POD_OPT_Int(&info->rate),
-			SPA_FORMAT_AUDIO_channels,      SPA_POD_OPT_Int(&info->channels),
-			SPA_FORMAT_AUDIO_position,      SPA_POD_OPT_Pod(&position));
-	if (position == NULL ||
-	    !spa_pod_copy_array(position, SPA_TYPE_Id, info->position, SPA_AUDIO_MAX_CHANNELS))
-		SPA_FLAG_SET(info->flags, SPA_AUDIO_FLAG_UNPOSITIONED);
-
-	return res;
-}
-
-int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, struct channel_map *map,
+int format_parse_param(const struct spa_pod *param, bool collect,
+		struct sample_spec *ss, struct channel_map *map,
 		const struct sample_spec *def_ss, const struct channel_map *def_map)
 {
 	struct spa_audio_info info = { 0 };
@@ -457,28 +523,50 @@ int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, stru
 
 	switch (info.media_subtype) {
 	case SPA_MEDIA_SUBTYPE_raw:
+		if (spa_format_audio_raw_parse(param, &info.info.raw) < 0)
+	                return -ENOTSUP;
 		if (def_ss != NULL) {
 			if (ss != NULL)
 				*ss = *def_ss;
-			if (audio_raw_parse_opt(param, &info.info.raw) < 0)
-		                return -ENOTSUP;
 		} else {
-			if (spa_format_audio_raw_parse(param, &info.info.raw) < 0)
-		                return -ENOTSUP;
+			if (info.info.raw.rate == 0)
+				info.info.raw.rate = 48000;
+			if (info.info.raw.format == 0 ||
+			    info.info.raw.rate == 0 ||
+			    info.info.raw.channels == 0 ||
+			    info.info.raw.channels > SPA_AUDIO_MAX_CHANNELS)
+				return -ENOTSUP;
 		}
 		break;
 	case SPA_MEDIA_SUBTYPE_iec958:
 	{
 		struct spa_audio_info_iec958 iec;
 
+		if (collect)
+			break;
+
 		if (spa_format_audio_iec958_parse(param, &iec) < 0)
 			return -ENOTSUP;
 
 		info.info.raw.format = SPA_AUDIO_FORMAT_S16;
-		info.info.raw.channels = 2;
 		info.info.raw.rate = iec.rate;
 		info.info.raw.position[0] = SPA_AUDIO_CHANNEL_FL;
 		info.info.raw.position[1] = SPA_AUDIO_CHANNEL_FR;
+		switch (iec.codec) {
+		case SPA_AUDIO_IEC958_CODEC_TRUEHD:
+		case SPA_AUDIO_IEC958_CODEC_DTSHD:
+			info.info.raw.channels = 8;
+			info.info.raw.position[2] = SPA_AUDIO_CHANNEL_FC;
+			info.info.raw.position[3] = SPA_AUDIO_CHANNEL_LFE;
+			info.info.raw.position[4] = SPA_AUDIO_CHANNEL_SL;
+			info.info.raw.position[5] = SPA_AUDIO_CHANNEL_SR;
+			info.info.raw.position[6] = SPA_AUDIO_CHANNEL_RL;
+			info.info.raw.position[7] = SPA_AUDIO_CHANNEL_RR;
+			break;
+		default:
+			info.info.raw.channels = 2;
+			break;
+		}
 		break;
 	}
 	default:
@@ -493,9 +581,11 @@ int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, stru
 		        ss->channels = info.info.raw.channels;
 	}
 	if (map) {
-		map->channels = info.info.raw.channels;
-		for (i = 0; i < map->channels; i++)
-			map->map[i] = info.info.raw.position[i];
+		if (info.info.raw.channels) {
+			map->channels = info.info.raw.channels;
+			for (i = 0; i < map->channels; i++)
+				map->map[i] = info.info.raw.position[i];
+		}
 	}
 	return 0;
 }
@@ -597,6 +687,9 @@ static int add_int(struct format_info *info, const char *k, struct spa_pod *para
 	if (val->type != SPA_TYPE_Int)
 		return -ENOTSUP;
 
+	if (n_values == 0)
+		return -ENOENT;
+
 	values = SPA_POD_BODY(val);
 
 	switch (choice) {
@@ -613,7 +706,9 @@ static int add_int(struct format_info *info, const char *k, struct spa_pod *para
 		size_t size;
 		FILE *f;
 
-		f = open_memstream(&ptr, &size);
+		if ((f = open_memstream(&ptr, &size)) == NULL)
+			return -errno;
+
 		fprintf(f, "[");
 		for (i = 1; i < n_values; i++)
 			fprintf(f, "%s %d", i == 1 ? "" : ",", values[i]);
@@ -773,7 +868,7 @@ int format_info_to_spec(const struct format_info *info, struct sample_spec *ss,
 	if (spa_json_is_float(val, len)) {
 		if (spa_json_parse_float(val, len, &f) <= 0)
 			return -EINVAL;
-		ss->channels = f;
+		ss->channels = (uint8_t)f;
 	} else if (spa_json_is_array(val, len)) {
 		return -ENOTSUP;
 	} else if (spa_json_is_object(val, len)) {
